# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kogito',
 'kogito.core',
 'kogito.core.processors',
 'kogito.core.processors.models',
 'kogito.evaluation',
 'kogito.evaluation.bert_score',
 'kogito.evaluation.bleu',
 'kogito.evaluation.cider',
 'kogito.evaluation.meteor',
 'kogito.evaluation.rouge',
 'kogito.linkers',
 'kogito.models',
 'kogito.models.bart',
 'kogito.models.gpt2',
 'kogito.models.gpt3']

package_data = \
{'': ['*'], 'kogito.core.processors': ['data/*']}

install_requires = \
['bert-score>=0.3.11,<0.4.0',
 'grpcio>=1.51.1,<1.52.0',
 'inflect>=5.3.0,<6.0.0',
 'openai>=0.18.1,<0.19.0',
 'pandas>=1.3.5,<2.0.0',
 'pytorch-lightning>=1.5.10,<1.6.0',
 'rouge-score>=0.0.4,<0.0.5',
 'sacrebleu>=2.0.0,<3.0.0',
 'sentencepiece>=0.1.97,<0.2.0',
 'spacy>=3.2.3,<4.0.0',
 'torch>=1.10.1,<2.0.0',
 'transformers>=4.15.0,<5.0.0',
 'wandb>=0.12.9,<0.13.0']

setup_kwargs = {
    'name': 'kogito',
    'version': '0.6.0',
    'description': 'A Python NLP Commonsense Knowledge Inference Toolkit',
    'long_description': '# kogito\nA Python NLP Commonsense Knowledge Inference Tool\n\nSystem Description available here: https://arxiv.org/abs/2211.08451\n\n## Installation\n\n### Installation with pip\n**kogito** can be installed using pip.\n\n```sh\npip install kogito\n```\n\nIt requires a minimum ``python`` version of ``3.8``.\n\n## Setup\n\n### Inference\n**kogito** uses [spacy](https://spacy.io) under the hood for various text processing purposes, so, a [spacy](https://spacy.io) language package has to be installed before running the inference module.\n\n```sh\npython -m spacy download en_core_web_sm\n``` \nBy default, ``CommonsenseInference`` module uses ``en_core_web_sm`` to initialize ``spacy`` pipeline, but a different language pipeline can be specified as well.\n\n### Evaluation\nIf you also would like evaluate knowledge models using `METEOR` score, then you need to download the following ``nltk`` libraries:\n```python\nimport nltk\n\nnltk.download("punkt")\nnltk.download("wordnet")\nnltk.download("omw-1.4")\n```\n\n## Quickstart\n**kogito** provides an easy interface to interact with knowledge inference or commonsense reasoning models such as [COMET](https://arxiv.org/abs/2010.05953) to generate inferences from a text input.\nHere is a sample usage of the library where you can initialize an inference module, a custom commonsense reasoning model, and generate a knowledge graph from text on the fly.\n\n```python\nfrom kogito.models.bart.comet import COMETBART\nfrom kogito.inference import CommonsenseInference\n\n# Load pre-trained model from HuggingFace\nmodel = COMETBART.from_pretrained("mismayil/comet-bart-ai2")\n\n# Initialize inference module with a spacy language pipeline\ncsi = CommonsenseInference(language="en_core_web_sm")\n\n# Run inference\ntext = "PersonX becomes a great basketball player"\nkgraph = csi.infer(text, model)\n\n# Save output knowledge graph to JSON file\nkgraph.to_jsonl("kgraph.json")\n```\n\nHere is an excerpt from the result of the above code sample:\n\n```json\n{"head": "PersonX becomes a great basketball player", "relation": "Causes", "tails": [" PersonX practices every day.", " PersonX plays basketball every day", " PersonX practices every day"]}\n{"head": "basketball", "relation": "ObjectUse", "tails": [" play with friends", " play basketball with", " play basketball"]}\n{"head": "player", "relation": "CapableOf", "tails": [" play game", " win game", " play football"]}\n{"head": "great basketball player", "relation": "HasProperty", "tails": [" good at basketball", " good at sports", " very good"]}\n{"head": "become player", "relation": "isAfter", "tails": [" play game", " become coach", " play with"]}\n```\nThis is just one way to generate commonsense inferences and **kogito** offers much more. For complete documentation, check out the [kogito docs](https://kogito.readthedocs.io).\n\n## Development\n\n### Setup\n**kogito** uses [Poetry](https://python-poetry.org/) to manage its dependencies. \n\nInstall poetry from the official repository first:\n```sh\ncurl -sSL https://install.python-poetry.org | python3 -\n```\n\nThen run the following command to install package dependencies:\n```sh\npoetry install\n```\n\n## Data\nIf you need the ATOMIC2020 data to train your knowledge models, you can download it from AI2:\n\nFor ATOMIC:\n```sh\nwget https://storage.googleapis.com/ai2-mosaic/public/atomic/v1.0/atomic_data.tgz\n```\n\nFor ATOMIC 2020:\n```sh\nwget https://ai2-atomic.s3-us-west-2.amazonaws.com/data/atomic2020_data-feb2021.zip\n```\n\n## Paper\nIf you want to learn more about the library design, models and data used for this toolkit, check out our [paper](https://arxiv.org/abs/2211.08451). The paper can be cited as:\n\n```\n@article{Ismayilzada2022kogito,\n  title={kogito: A Commonsense Knowledge Inference Toolkit},\n  author={Mete Ismayilzada and Antoine Bosselut},\n  journal={ArXiv},\n  volume={abs/2211.08451},\n  year={2022}\n}\n```\n\nIf you work with knowledge models, consider citing the following papers:\n\n```\n@article{Hwang2020COMETATOMIC,\n author = {Jena D. Hwang and Chandra Bhagavatula and Ronan Le Bras and Jeff Da and Keisuke Sakaguchi and Antoine Bosselut and Yejin Choi},\n booktitle = {Proceedings of the 35th AAAI Conference on Artificial Intelligence (AAAI)},\n title = {COMET-ATOMIC 2020: On Symbolic and Neural Commonsense Knowledge Graphs},\n year = {2021}\n}\n\n@inproceedings{Bosselut2019COMETCT,\n author = {Antoine Bosselut and Hannah Rashkin and Maarten Sap and Chaitanya Malaviya and Asli Çelikyilmaz and Yejin Choi},\n booktitle = {Proceedings of the 57th Annual Meeting of the Association for Computational Linguistics (ACL)},\n title = {COMET: Commonsense Transformers for Automatic Knowledge Graph Construction},\n year = {2019}\n}\n```\n\n## Acknowledgements\nSignificant portion of the model training and evaluation code has been adapted from the original [codebase](https://github.com/allenai/comet-atomic-2020) for the paper [(Comet-) Atomic 2020: On Symbolic and Neural Commonsense Knowledge Graphs.](https://www.semanticscholar.org/paper/COMET-ATOMIC-2020%3A-On-Symbolic-and-Neural-Knowledge-Hwang-Bhagavatula/e39503e01ebb108c6773948a24ca798cd444eb62)\n',
    'author': 'Mete Ismayil',
    'author_email': 'mahammad.ismayilzada@epfl.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/epfl-nlp/kogito',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from pathlib import Path
from typing import Dict
import warnings
import torch
from torch.utils.data import Dataset

from kogito.core.knowledge import GEN_TOKEN, EOS_TOKEN
from kogito.core.utils import encode_line, trim_batch, SortishSampler


class KnowledgeDataset(Dataset):
    def __init__(self, kg_graph, tokenizer, source_len, target_len, is_eval=False):
        self.tokenizer = tokenizer
        self.data = kg_graph.to_dataframe()
        self.source_len = source_len
        self.target_len = target_len
        self.text = self.data["head"] + " " + self.data["relation"] + f" {GEN_TOKEN}"
        self.ctext = self.data["tails"] + f" {EOS_TOKEN}"
        self.is_eval = is_eval

    def __len__(self):
        return len(self.text)

    def __getitem__(self, index):
        text = str(self.text[index])
        text = " ".join(text.split())

        ctext = str(self.ctext[index])
        ctext = " ".join(ctext.split())

        if self.is_eval:
            source = self.tokenizer.batch_encode_plus(
                [text],
                pad_to_max_length=False,
                max_length=self.source_len,
                return_tensors="pt",
                truncation=True,
            )
            target = self.tokenizer.batch_encode_plus(
                [ctext],
                pad_to_max_length=False,
                max_length=self.target_len,
                return_tensors="pt",
                truncation=True,
            )
        else:
            source = self.tokenizer.batch_encode_plus(
                [text + " " + ctext],
                pad_to_max_length=True,
                max_length=self.source_len + self.target_len,
                return_tensors="pt",
                truncation=True,
            )
            target = source

        source_ids = source["input_ids"].squeeze()
        source_mask = source["attention_mask"].squeeze()
        target_ids = target["input_ids"].squeeze()
        target_mask = target["attention_mask"].squeeze()

        return {
            "source_ids": source_ids.to(dtype=torch.long),
            "source_mask": source_mask.to(dtype=torch.long),
            "target_ids": target_ids.to(dtype=torch.long),
            "target_mask": target_mask.to(dtype=torch.long),
        }


class Seq2SeqDataset(Dataset):
    def __init__(
        self,
        tokenizer,
        train_graph,
        max_source_length,
        max_target_length,
        type_path="train",
        n_obs=None,
        src_lang=None,
        tgt_lang=None,
        prefix="",
        val_graph=None,
        test_graph=None,
    ):
        super().__init__()
        self.train_graph = train_graph
        self.val_graph = val_graph
        self.test_graph = test_graph
        self.input_graph = self.train_graph

        if type_path == "val":
            self.input_graph = self.val_graph

        if type_path == "test":
            self.input_graph = self.test_graph

        self.src_lens = [
            len(f"{kg.head} {kg.relation} {GEN_TOKEN}") for kg in self.input_graph
        ]
        self.max_source_length = max_source_length
        self.max_target_length = max_target_length
        self.tokenizer = tokenizer
        self.prefix = prefix
        if n_obs is not None:
            self.src_lens = self.src_lens[:n_obs]
        self.pad_token_id = self.tokenizer.pad_token_id
        self.src_lang = src_lang
        self.tgt_lang = tgt_lang

    def __len__(self):
        return len(self.src_lens)

    def __getitem__(self, index) -> Dict[str, torch.Tensor]:
        kg = self.input_graph[index]
        source_line = self.prefix + f"{kg.head} {kg.relation} {GEN_TOKEN}"
        tgt_line = kg.tails[0] if kg.tails else None
        assert source_line, f"empty source line for index {index}"
        assert tgt_line, f"empty tgt line for index {index}"
        source_inputs = encode_line(self.tokenizer, source_line, self.max_source_length)
        target_inputs = encode_line(self.tokenizer, tgt_line, self.max_target_length)

        source_ids = source_inputs["input_ids"].squeeze()
        target_ids = target_inputs["input_ids"].squeeze()
        src_mask = source_inputs["attention_mask"].squeeze()
        return {
            "input_ids": source_ids,
            "attention_mask": src_mask,
            "decoder_input_ids": target_ids,
        }

    @staticmethod
    def get_char_lens(data_file):
        return [len(x) for x in Path(data_file).open().readlines()]

    @staticmethod
    def trim_seq2seq_batch(batch, pad_token_id) -> tuple:
        y = trim_batch(batch["decoder_input_ids"], pad_token_id)
        source_ids, source_mask = trim_batch(
            batch["input_ids"], pad_token_id, attention_mask=batch["attention_mask"]
        )
        return source_ids, source_mask, y

    def collate_fn(self, batch) -> Dict[str, torch.Tensor]:
        input_ids = torch.stack([x["input_ids"] for x in batch])
        masks = torch.stack([x["attention_mask"] for x in batch])
        target_ids = torch.stack([x["decoder_input_ids"] for x in batch])
        pad_token_id = self.pad_token_id
        y = trim_batch(target_ids, pad_token_id)
        source_ids, source_mask = trim_batch(
            input_ids, pad_token_id, attention_mask=masks
        )
        batch = {
            "input_ids": source_ids,
            "attention_mask": source_mask,
            "decoder_input_ids": y,
        }
        return batch

    def make_sortish_sampler(self, batch_size):
        return SortishSampler(self.src_lens, batch_size)


class MBartDataset(Seq2SeqDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.max_source_length != self.max_target_length:
            warnings.warn(
                f"Mbart will ignore max_target_length = {self.max_target_length} and"
                "use {self.max_source_length} for both sides."
            )

    def __getitem__(self, index) -> Dict[str, str]:
        kg = self.input_graph[index]
        source_line = self.prefix + f"{kg.head} {kg.relation} {GEN_TOKEN}"
        tgt_line = kg.tails[0] if kg.tails else None
        assert source_line, f"empty source line for index {index}"
        assert tgt_line, f"empty tgt line for index {index}"
        return {
            "tgt_texts": source_line,
            "src_texts": tgt_line,
        }

    def collate_fn(self, batch) -> Dict[str, torch.Tensor]:
        batch_encoding = self.tokenizer.prepare_translation_batch(
            [x["src_texts"] for x in batch],
            src_lang=self.src_lang,
            tgt_texts=[x["tgt_texts"] for x in batch],
            tgt_lang=self.tgt_lang,
            max_length=self.max_source_length,
        )
        return batch_encoding.data



from kogito.evaluation.bert_score.score import score

# Code for BertScore reused from original implementation: https://github.com/Tiiiger/bert_score


class BertScore:
    def __init__(self):
        self._hypo_for_image = {}
        self.ref_for_image = {}

    def compute_score(self, gts, res):

        assert gts.keys() == res.keys()
        imgIds = gts.keys()

        hyp_input = []
        ref_input = []
        same_indices = []
        for id in imgIds:
            hypo = res[id]
            ref = gts[id]

            # Sanity check.
            assert type(hypo) is list
            assert len(hypo) == 1
            assert type(ref) is list
            assert len(ref) >= 1

            hyp_input += [hypo[0]] * len(ref)
            ref_input += ref
            same_indices.append(len(ref_input))

        p, r, f_scores = score(hyp_input, ref_input)

        prev_idx = 0
        aggreg_f1_scores = []
        for idx in same_indices:
            aggreg_f1_scores.append(f_scores[prev_idx:idx].mean().cpu().item())
            prev_idx = idx

        return sum(aggreg_f1_scores) / len(aggreg_f1_scores), aggreg_f1_scores

    def method(self):
        return "Bert Score"



from typing import List, Tuple

from kogito.evaluation.bleu.bleu import Bleu
from kogito.evaluation.meteor.meteor import Meteor
from kogito.evaluation.rouge.rouge import Rouge
from kogito.evaluation.cider.cider import Cider
from kogito.evaluation.bert_score.bert_score import BertScore
from kogito.core.knowledge import Knowledge


METRIC_MAP = {
    "bleu": (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
    "meteor": (Meteor(), "METEOR"),
    "rouge": (Rouge(), "ROUGE_L"),
    "cider": (Cider(), "CIDEr"),
    "bert-score": (BertScore(), "Bert Score"),
}


class Evaluator:
    def __init__(self, gts, res, metrics):
        self.gts = gts
        self.res = res
        self.scorers = [METRIC_MAP[metric] for metric in metrics]

    def evaluate(self):
        score_dict = {}

        for scorer, method in self.scorers:
            score, _ = scorer.compute_score(self.gts, self.res)
            if type(method) == list:
                for sc, m in zip(score, method):
                    score_dict[m] = str(sc)
            else:
                score_dict[method] = score

        return score_dict


def topk_eval(data: List[Tuple[Knowledge, List[str]]], metrics, k=1):
    topk_gts = {}
    topk_res = {}

    for i, (kg, reference) in enumerate(data):
        for j, g in enumerate(kg.tails[:k]):
            key = str(i) + "_" + str(j)
            topk_gts[key] = reference
            topk_res[key] = [g]

    evaluator = Evaluator(topk_gts, topk_res, metrics)
    scores = evaluator.evaluate()

    return scores



from typing import List

from abc import ABC, abstractmethod, abstractclassmethod
from kogito.core.knowledge import KnowledgeGraph
from kogito.evaluation.eval import topk_eval, METRIC_MAP


class KnowledgeModel(ABC):
    """
    Base class to represent a Knowledge Model.
    """

    @abstractmethod
    def train(self, train_graph: KnowledgeGraph, *args, **kwargs) -> "KnowledgeModel":
        """Train a knowledge model

        Args:
            train_graph (KnowledgeGraph): Training dataset

        Raises:
            NotImplementedError: This method has to be implemented
                                 by concrete subclasses.

        Returns:
            KnowledgeModel: Trained knowledge model
        """
        raise NotImplementedError

    @abstractmethod
    def generate(self, input_graph: KnowledgeGraph, *args, **kwargs) -> KnowledgeGraph:
        """Generate inferences from knowledge model

        Args:
            input_graph (KnowledgeGraph): Input dataset

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.

        Returns:
            KnowledgeGraph: Input graph with tails generated
        """
        raise NotImplementedError

    @abstractmethod
    def save_pretrained(self, save_path: str) -> None:
        """Save model as a pretrained model

        Args:
            save_path (str): Directory to save the model to.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.
        """
        raise NotImplementedError

    @abstractclassmethod
    def from_pretrained(cls, model_name_or_path: str) -> "KnowledgeModel":
        """Load model from a pretrained model path
        This method can load models either from HuggingFace by model name
        or from disk by model path.

        Args:
            model_name_or_path (str): HuggingFace model name or local model path to load from.

        Raises:
            NotImplementedError: This method has to be implemented by
                                 concrete subclasses.

        Returns:
            KnowledgeModel: Loaded knowledge model.
        """
        raise NotImplementedError

    def evaluate(
        self,
        input_graph: KnowledgeGraph,
        metrics: List[str] = ["bleu", "meteor", "rouge", "cider", "bert-score"],
        top_k: int = 1,
        *args,
        **kwargs,
    ) -> dict:
        """Evaluate model on various metrics.
        Input graph should contain the reference tails, so that it can be used to score the model generations
        on the same input graph. Any arguments provided aside from the ones accepted by this method will be
        passed onto the ``KnowledgeModel.generate`` method.

        Args:
            input_graph (KnowledgeGraph): Input graph to evaluate. Should contain the ground truth tails.
            metrics (List[str], optional): Metrics to compute.
                Defaults to ["bleu", "meteor", "rouge", "cider", "bert-score"].
            top_k (int, optional): Top k generations to evaluate. Defaults to 1.
            *args (optional): Extra arguments for `KnowledgeModel.generate` method.
            **kwargs (optional): Extra keyword arguments for ``KnowledgeModel.generate`` method.

        Returns:
            dict: Dictionary of scores
        """
        return evaluate(self, input_graph, metrics, top_k=top_k, *args, **kwargs)


def evaluate(
    model: KnowledgeModel,
    input_graph: KnowledgeGraph,
    metrics: List[str] = ["bleu", "meteor", "rouge", "cider", "bert-score"],
    top_k: int = 1,
    *args,
    **kwargs,
):
    if not set(metrics).issubset(set(METRIC_MAP.keys())):
        raise ValueError(
            f"Invalid evaluation metrics found: {set(metrics) - set(METRIC_MAP.keys())}"
        )

    output_graph = model.generate(input_graph=input_graph, *args, **kwargs)
    evaluation_data = []

    for input_kg, output_kg in zip(input_graph, output_graph):
        assert input_kg.head == output_kg.head
        assert input_kg.relation == output_kg.relation
        assert len(input_kg.tails) > 0

        evaluation_data.append((output_kg, input_kg.tails))

    return topk_eval(evaluation_data, metrics, k=top_k)



import torch
from collections import defaultdict
from transformers import BertTokenizer, BertModel

from kogito.evaluation.bert_score.utils import (
    get_idf_dict,
    bert_cos_score_idf,
    bert_types,
)


def score(
    cands,
    refs,
    bert="bert-base-multilingual-cased",
    num_layers=8,
    no_idf=False,
    batch_size=64,
):
    """
    BERTScore metric.
    Args:
        - :param: `cands` (list of str): candidate sentences
        - :param: `refs` (list of str): reference sentences
        - :param: `bert` (str): bert specification
        - :param: `num_layers` (int): the layer of representation to use
        - :param: `verbose` (bool): turn on intermediate status update
        - :param: `no_idf` (bool): do not use idf weighting
        - :param: `batch_size` (int): bert score processing batch size
    """
    assert len(cands) == len(refs)
    assert bert in bert_types

    tokenizer = BertTokenizer.from_pretrained(bert)
    model = BertModel.from_pretrained(bert)
    model.eval()
    device = "cuda" if torch.cuda.is_available() else "cpu"
    model.to(device)

    # drop unused layers
    model.encoder.layer = torch.nn.ModuleList(
        [layer for layer in model.encoder.layer[:num_layers]]
    )

    if no_idf:
        idf_dict = defaultdict(lambda: 1.0)
        # set idf for [SEP] and [CLS] to 0
        idf_dict[101] = 0
        idf_dict[102] = 0
    else:
        idf_dict = get_idf_dict(refs, tokenizer)

    all_preds = bert_cos_score_idf(
        model, refs, cands, tokenizer, idf_dict, device=device, batch_size=batch_size
    )

    P = all_preds[:, 0].cpu()
    R = all_preds[:, 1].cpu()
    F1 = all_preds[:, 2].cpu()

    return P, R, F1
